/* 
   - The contents of this file are subject to the Mozilla Public
   - License Version 1.1 (the "License"); you may not use this file
   - except in compliance with the License. You may obtain a copy of
   - the License at http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS
   - IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
   - implied. See the License for the specific language governing
   - rights and limitations under the License.
   -
   - The Original Code is this file as it was released on
   - March 20, 2001.
   -
   - The Initial Developer of the Original Code is Henk-Johan van Rantwijk.
   - Portions created by Henk-Johan van Rantwijk are Copyright (C) 2001
   - Henk-Johan van Rantwijk.  All Rights Reserved.
   -
   - Contributor(s):
   -
   - Alternatively, the contents of this file may be used under the
   - terms of the GNU General Public License Version 2 or later (the
   - "GPL"), in which case the provisions of the GPL are applicable
   - instead of those above.  If you wish to allow use of your
   - version of this file only under the terms of the GPL and not to
   - allow others to use your version of this file under the MPL,
   - indicate your decision by deleting the provisions above and
   - replace them with the notice and other provisions required by
   - the GPL.  If you do not delete the provisions above, a recipient
   - may use your version of this file under either the MPL or the
   - GPL.
   -
*/

const gDebug = false; // change this to true to enable debug info on the console

var mzRDFStore = null; // global pointer to our datasource (multizilla.rdf)

var RDF = Components.classes[ "@mozilla.org/rdf/rdf-service;1" ]
                    .getService( Components.interfaces.nsIRDFService );

var RDFCUtils = Components.classes[ "@mozilla.org/rdf/container-utils;1" ]
                          .getService( Components.interfaces.nsIRDFContainerUtils );

/**
 * File mode bits
 * #define PR_IRWXU 00700  read, write, execute/search by owner
 * #define PR_IRUSR 00400  read permission, owner
 * #define PR_IWUSR 00200  write permission, owner
 * #define PR_IXUSR 00100  execute/search permission, owner
 * #define PR_IRWXG 00070  read, write, execute/search by group
 * #define PR_IRGRP 00040  read permission, group
 * #define PR_IWGRP 00020  write permission, group
 * #define PR_IXGRP 00010  execute/search permission, group
 * #define PR_IRWXO 00007  read, write, execute/search by others
 * #define PR_IROTH 00004  read permission, others
 * #define PR_IWOTH 00002  write permission, others
 * #define PR_IXOTH 00001  execute/search permission, others
 * 
 * File open bits
 * #define PR_RDONLY       0x01
 * #define PR_WRONLY       0x02
 * #define PR_RDWR         0x04
 * #define PR_CREATE_FILE  0x08
 * #define PR_APPEND       0x10
 * #define PR_TRUNCATE     0x20
 * #define PR_SYNC         0x40
 * #define PR_EXCL         0x80
*/

//////////////////////////////////////////////////////////////////////////
// This function returns a nsiURI of the users profile directory or null 
// in case of failure. It is used in getNewOrCurrentDSByFilename()
//////////////////////////////////////////////////////////////////////////
function getProfileURI()
{
  var profile = Components.classes["@mozilla.org/profile/manager;1"].getService(Components.interfaces.nsIProfileInternal);
  var gIOService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
  var URI = gIOService.newFileURI(profile.getProfileDir(profile.currentProfile));
  return URI || null;
}
////////////////////////////////////////////////////////////////////////////////
// This initializes or build a new RDF
////////////////////////////////////////////////////////////////////////////////
function initializeRDFStorageFile() 
{
  mzRDFStore = getNewOrCurrentDSByFilename("multizilla.rdf", true, null)  // "quickprefs.rdf" for stand-alone versions of QPM
}
////////////////////////////////////////////////////////////////////////////////
// This initializes or build a new RDF (multizilla.rdf) It depends on two other 
// function, getProfileURI() and mzWriteToFile() so keep that in mind. It 
// returns the datasource or null on failure
////////////////////////////////////////////////////////////////////////////////
function getNewOrCurrentDSByFilename(aFileName, aCreateWhenMissing, aData)
{
  var rdfFile = aFileName;
  // Get profile directory of current profile
  var profileURI = getProfileURI();
  // Construct file handler to our RDF storage file
  const ioService = Components.classes['@mozilla.org/network/io-service;1']
                              .getService(Components.interfaces.nsIIOService);
  var gFileHandler = ioService.getProtocolHandler("file")
                              .QueryInterface(Components.interfaces.nsIFileProtocolHandler);
  var storageFile = gFileHandler.getFileFromURLSpec(profileURI.spec + rdfFile)
                                .QueryInterface(Components.interfaces.nsILocalFile);
  // Do we need to create a new file?
  if (aCreateWhenMissing && !storageFile.exists()) {
    /* The RDF (aFileName) is missing, we need to construct one ourself as fallback
       Note: this has partly been ported from C++ (nsLocalStore.cpp#410) */
    const defaultRDF = "<?xml version=\"1.0\"?>\n" +
                       "<RDF:RDF xmlns:RDF=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"\n" +
                       "         xmlns:MZ=\"http://multizilla.mozdev.org/rdf#\">\n" +
                       "  <!-- Empty -->\n" +
                       "</RDF:RDF>\n";
    if (aData) // use specified data
      defaultRDF = aData;
    // create 'storageFile' and store 'defaultRDF' in it
    mzWriteToFile(storageFile, defaultRDF);
  }
  return RDF.GetDataSource(profileURI.spec + rdfFile) || null;
}
////////////////////////////////////////////////////////////////////////////////
// This function returns a RDFContainer based on the given datasource 
// 'aDatasource', the given type 'aType' and the given resource 'aResource'. It 
// will creates a new container when you call this function with 'true' for 
// 'aInit'. It returns null on failure.
////////////////////////////////////////////////////////////////////////////////
function mzConstructRDFContainer(aDatasource, aType, aResource)
{
  var container = Components.classes["@mozilla.org/rdf/container;1"]
                            .createInstance( Components.interfaces.nsIRDFContainer );

  if (aDatasource && aType && aResource) {
    try {
      switch(aType) {
        default:
        case "seq": RDFCUtils.MakeSeq(aDatasource, RDF.GetResource(aResource));
          break;
        case "bag": RDFCUtils.MakeBag(aDatasource, RDF.GetResource(aResource));
          break;
        case "alt": RDFCUtils.MakeAlt(aDatasource, RDF.GetResource(aResource));
          break;
      }
      container.Init(aDatasource, RDF.GetResource(aResource));
    } catch (ex) {
      mzDebug(ex);
      return null;
    }
  }
  return container;
}
////////////////////////////////////////////////////////////////////////////////
function saveBrowserSession()
{
  // Was the browser startup for an external link/external application?
  if (!gBrowser.mRestoreBrowserSession) {
    var mzBundle = srGetStrBundle("chrome://multiviews/locale/multiviews.properties");
    var checkboxMsg = mzBundle.GetStringFromName("saveBrowserSessionDontAskAgain");
    var title = mzBundle.GetStringFromName("saveBrowserSessionTitle");
    var text = mzBundle.GetStringFromName("saveBrowserSessionText");
    var checkbox = {value:0};
    var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    var result = promptService.confirmEx(window, title, text, 
                                        (promptService.BUTTON_POS_0 * promptService.BUTTON_TITLE_DONT_SAVE) +
                                        (promptService.BUTTON_POS_1 * promptService.BUTTON_TITLE_SAVE),
                                         null, null, null, null, checkbox);
     if (result == 0) // Don't save this session
      return;
  }
  var sessionID = 0;
  var windowID = 0;
  /***
  * This can easily be expanded to open all tabs of all windows, but that
  * will have to be introduced in a next update because of lack of time (:
  */
  storeGroupData("urn:session-"+sessionID+":window-"+windowID+":tabs");
}
////////////////////////////////////////////////////////////////////////////////
function storeGroupData(aResource)
{
  var tab = gBrowser.mTabList;
  removeAllResourcesFor(aResource + ":root"); // "urn:session-x:window-x:tabs" + ":root"
  var tabContainer = mzConstructRDFContainer(mzRDFStore, "seq", aResource + ":root");

  if (!tabContainer)
    return false;

  var tabs = gBrowser.mTabContainer.childNodes.length;
  var propertyNames = ["allowPlugins", "allowJavascript", "allowMetaRedirects", "allowSubframes", "allowImages"];

  for (var index = 0; index < tabs; index++) {
    /* Collect the Tabs history
    var tabHistory = gBrowser.getTabHistory(tab[index]);
    for (var i in tabHistory.tabHistoryEntries) {
      // History information substracted from the tabs history
      var uri = tabHistory.tabHistoryEntries[i].uri;
      var title = tabHistory.tabHistoryEntries[i].title;
      var isSubFrame = tabHistory.tabHistoryEntries[i].isSubFrame;
      var x = tabHistory.tabHistoryEntries[i].x;
      var y = tabHistory.tabHistoryEntries[i].y;
      dump("\n\nuri: "+uri+"\ntitle: "+title+"\nisSubFrame: "+isSubFrame+"\nx: "+x+"\ny: "+y);
    }
    dump("\nTab: "+i+"tabHistory.tabHistoryIndex: "+tabHistory.tabHistoryIndex); */

    var browser = gBrowser.getBrowserForTab(tab[index]);
    var location = browser.webNavigation.currentURI.spec;
    if (location == "about:blank") // skip unused (blank) tabs
      continue;
    var label = tab[index].getAttribute("label");
    var customLabel = tab[index].getAttribute("customLabel") ? tab[index].getAttribute("label") : "false";

    var image = tab[index].getAttribute("image");
    var interval = tab[index].mReloadInterval;
    var ordinal = tab[index].ordinal;

    var selected = (tab[index] == gBrowser.selectedTab) ? "true" : "false";

    var newNode = RDF.GetResource(aResource + ":tab-" + index); // "urn:session-x:window-x:tabs" + ":tab-x"
    var MZ ="http://multizilla.mozdev.org/rdf#";
    var labelProp = RDF.GetResource(MZ+"label");
    var customProp = RDF.GetResource(MZ+"custom");
    var imageProp = RDF.GetResource(MZ+"image");
    var locationProp = RDF.GetResource(MZ+"location");
    var intervalProp = RDF.GetResource(MZ+"interval");
    var ordinalProp = RDF.GetResource(MZ+"ordinal");

    var selectedProp = RDF.GetResource(MZ+"selected");

    mzRDFStore.Assert(newNode, labelProp, RDF.GetLiteral(label), true);
    mzRDFStore.Assert(newNode, customProp, RDF.GetLiteral(customLabel), true);
    mzRDFStore.Assert(newNode, imageProp, RDF.GetLiteral(image), true);
    mzRDFStore.Assert(newNode, locationProp, RDF.GetLiteral(location), true);
    mzRDFStore.Assert(newNode, intervalProp, RDF.GetLiteral(interval), true);
    mzRDFStore.Assert(newNode, ordinalProp, RDF.GetLiteral(ordinal), true);
    // Store browser properties
    for (i in propertyNames) {
      var propertyValue = browser.docShell[propertyNames[i]].valueOf();
      var property = RDF.GetResource(MZ+propertyNames[i]);
      mzRDFStore.Assert(newNode, property, RDF.GetLiteral(propertyValue), true);
    }
    mzRDFStore.Assert(newNode, selectedProp, RDF.GetLiteral(selected), true);
    tabContainer.AppendElement(newNode, true);
    mzRDFStore.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
  }
  /***
  * This pref setting will trigger 'gTabSessionStartupPrefListener' so that
  * other pref settings will be changed accordantly.
  *
  * var shouldSave = readMyPref("multizilla.tabs.saveonexit", "bool");
  * writePref("bool", "multizilla.tabs.startgroup", true);
  *
  * I removed the pref. This gives the user a choice, more control of this feature. 
  * The end user might want to save tabs on exit, for whatever reason, but that doesn't 
  * grant us the right to reload them on startup. We should ask his permission first!
  */
}
////////////////////////////////////////////////////////////////////////////////
// Modified function, from a prior version, restores all tabs/windows
function restoreBrowserSession(aCheckFirstWindow)
{
  if ((!isFirstWindow() && aCheckFirstWindow) || !mzRDFStore)
    return;

  // We can only check for duplicates *after* the startup page(s) is/are loaded!
  if (readMyPref("multizilla.prevent.double", "bool", true) || 
      readMyPref("browser.startup.page", "int", 0) > 0) {
    var delayed = false;
    var tab = null;
    if (pref.getIntPref("browser.startup.page") == 1) {
      var pageCount = pref.getIntPref("browser.startup.homepage.count");
      tab = gBrowser.mTabContainer.childNodes;
      if (pageCount != tab.length)
        delayed = true;
      for (var i = 0; i < tab.length; i++) {
        if (tab[i].hasAttribute("busy")) {
          delayed = true;
          break;
        }
      }
    }
    if (delayed && aCheckFirstWindow) {
      setTimeout(restoreBrowserSession, 1000, true);
      return;
    }
  }
  // Lets restore all tabs from a priorly saved browser session
  restoreTabsFromResource("urn:session-0:window-0:tabs:root");
}
//////////////////////////////////////////////////////////////////////////
function restoreTabsFromResource(aResource)
{
  var tabContainer = RDFCUtils.MakeSeq(mzRDFStore, RDF.GetResource(aResource)).GetElements();

  if (!tabContainer)
    return false;

  var index = 0;
  var addTab = true; // default is to add new tabs
  var useBlankTabs = readMyPref("multizilla.tabs.blank-tab-first", "bool", true);
  // Should we use new (blank) tabs before opening new tabs?
  if (useBlankTabs) {
    /*** 
    * The user wants to use unused (blank) tabs before opening new tabs. We *must* 
    * check the preference to find out if mozilla had to load one or more pages.
    * We can't simply use the first unused (blank) tab if that's the case!
    */	
    addTab = readMyPref("browser.startup.page", "int", 0) > 0 ? true : false;
  }
  // addTab is now either true or false, based on the users pref settings
  while (tabContainer.hasMoreElements()) {
    try {
      var resource = tabContainer.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
      var labels = mzRDFStore.ArcLabelsOut(resource);
      var dataCollection = new Array;
      var labelIndex = -1;
      while (labels.hasMoreElements()) {
        try {
          var label = labels.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
          var data = mzRDFStore.GetTarget(resource, label, true);
          var collection = {};
          collection.label = label.Value.replace(/.*#/, "") || "error";
          data = data.QueryInterface(Components.interfaces.nsIRDFLiteral);
          collection.value = data.Value;
          labelIndex++;
          dataCollection[labelIndex] = collection;
        } 
        catch(ex) { 
          mzDebug(ex);
        }
	  }
      index++;
      if (labelIndex >= 0)
        addTabFromCollection(dataCollection, (index > 1) ? true : addTab);
    } catch(ex) {
      mzDebug(ex);
      continue;
    }
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////
// This function opens a new tab, execept for the first URL, and updates 
// the corresponding browser properties for this tab. Some of the tabs 
// attributes are also changed, except for the ordinal. We saved the tabs 
// in the right order, so there's no need to set the ordinal
//////////////////////////////////////////////////////////////////////////
function addTabFromCollection(aDataCollection, newTabFlag)
{
  var URL = aDataCollection[3].value;
  /***
  * We don't want to open useless blank, ever.
  * We might also need to prevent duplicated tabs, but that's pref driven.
  */
  if (URL == "about:blank" || preventDupeTab(URL, true))
    return;
  var aTab = gBrowser.selectedTab;
  if (newTabFlag)
    aTab = gBrowser.addTab();
  // Get the browser for this (new) tab
  var browser = gBrowser.getBrowserForTab(aTab);
  // Modify browser properties before we load the URL for this tab
  browser.docShell.allowPlugins = (aDataCollection[6].value == "true") ? true : false;
  browser.docShell.allowJavascript = (aDataCollection[7].value == "true") ? true : false;
  browser.docShell.allowMetaRedirects = (aDataCollection[8].value == "true") ? true : false;
  browser.docShell.allowSubframes = (aDataCollection[9].value == "true") ? true : false;
  browser.docShell.allowImages = (aDataCollection[10].value == "true") ? true : false;
  // The properties are set so lets load the URL
  browser.loadURI(aDataCollection[3].value);
  // Do we need to set a reload interval?
  var interval = aDataCollection[4].value;
  // We only honor intervals of 30+ seconds
  if (interval >= 30000)
    gBrowser.setReloadInterval(aTab, interval);
  // Activate the (new) tab
  if (aDataCollection[11].value == "true")
    gBrowser.selectedTab = aTab;
}
//////////////////////////////////////////////////////////////////////////
// This function removes all given resources from multizilla.rdf
//////////////////////////////////////////////////////////////////////////
function removeAllResourcesFor(aResource)
{
  removeResources(aResource);
  removeDataResources(aResource);
  mzRDFStore.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
}
//////////////////////////////////////////////////////////////////////////
function removeResources(aResource)
{
  var tabContainer = RDFCUtils.MakeSeq(mzRDFStore, RDF.GetResource(aResource)).GetElements();

  if (!tabContainer)
    return false;

  /*** 
  * EXAMPLE: removeResources('urn:session-0:window-0:tabs:tab-0');
  * removes:
    <RDF:Description about="urn:session-0:window-0:tabs:tab-0"
                     MZ:label="mozdev.org - multizilla: index"
                     MZ:custom="false"
                     MZ:image="http://www.mozilla.org/images/mozilla-16.png"
                     MZ:location="http://multizilla.mozdev.org/index.html"
                     MZ:interval="0"
                     MZ:ordinal="0"
                     MZ:plugins="true"
                     MZ:javascript="true"
                     MZ:metaredirects="true"
                     MZ:subframes="true"
                     MZ:images="true"
                     MZ:selected="false" /> */

  while (tabContainer.hasMoreElements()) {
    try {
      var resource = tabContainer.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
      var labels = mzRDFStore.ArcLabelsOut(resource);

      while (labels.hasMoreElements()) {
        try {
          var label = labels.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
          var data = mzRDFStore.GetTarget(resource, label, true);
          mzRDFStore.Unassert(resource, label, data);
        } 
        catch(ex) { 
          mzDebug(ex);
        }
	  }
    } catch(ex) {
      mzDebug(ex);
      continue;
    }
  }
}
//////////////////////////////////////////////////////////////////////////
// The given resource will be removed from multizilla.rdf
function removeDataResources(aResource)
{
  var resource = RDF.GetResource(aResource);
  var labels = mzRDFStore.ArcLabelsOut(resource);
  /*** 
  * EXAMPLE: removeDataResources('urn:session-0:window-0:tabs:root');
  * removes:
    <RDF:Seq about="urn:session-0:window-0:tabs:root">
      <RDF:li resource="urn:session-0:window-0:tabs:tab-0"/>
    </RDF:Seq> */

  while (labels.hasMoreElements()) {
    try {
      var label = labels.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
      var value = mzRDFStore.GetTarget(resource, label, true);
      mzRDFStore.Unassert(resource, label, value);
    }
    catch(ex) {
      mzDebug(ex);
    }
  }
}
////////////////////////////////////////////////////////////////////////////////
// This function reads from a given file (aFile) and returns a pointer the 
// content or null on failure
////////////////////////////////////////////////////////////////////////////////	
function mzReadFromFile(aFile) 
{
  if (aFile) {
    var inputStream = Components.classes["@mozilla.org/network/file-input-stream;1"]
                                .createInstance(Components.interfaces.nsIFileInputStream);

    var scriptableStream = Components.classes["@mozilla.org/scriptableinputstream;1"]
                                     .createInstance(Components.interfaces.nsIScriptableInputStream);

    try {
      /**
       * inputStream.init( file, ioFlags, perm, behaviorFlags );
       *
       * @param file          - file to read from (must QI to nsILocalFile)
       * @param ioFlags       - file open flags (see File open bits)
       * @param permission    - file mode bits (see File mode bits)
       * @param behaviorFlags - flags specifying various behaviors of the class (see enumerations in the class)
      */
      inputStream.init(aFile, 1, 0, 0);
      scriptableStream.init(inputStream);
      var content = scriptableStream.read(scriptableStream.available());
      scriptableStream.close();
      inputStream.close();
    } catch(ex) {
      mzDebug(ex);
      return null;
    }
    return content;
  }
  return null;
}
////////////////////////////////////////////////////////////////////////////////
// This function write the given content (aContent) to a given file (aFile) and 
// it returns false on failure and true on success.
////////////////////////////////////////////////////////////////////////////////
function mzWriteToFile(aFile, aContent)
{
  if (aFile && aContent) {
    if (aFile.exists()) 
      aFile.remove(true);

    var outputStream = Components.classes["@mozilla.org/network/file-output-stream;1"]
                                 .createInstance(Components.interfaces.nsIFileOutputStream);

    try {
      aFile.create(aFile.NORMAL_FILE_TYPE, 0666);
      /**
       * outputStream.init( file, ioFlags, perm, behaviorFlags );
       *
       * @param file          - file to write to (must QI to nsILocalFile)
       * @param ioFlags       - file open flags (see File open bits)
       * @param permission    - file mode bits (see File mode bits)
       * @param behaviorFlags - currently none supported
      */
      outputStream.init(aFile, 2, 0x200, 0);
      outputStream.write(aContent, aContent.length);
      outputStream.close();
    } catch(ex) {
      mzDebug(ex);
      return false;
    }
    return true;
  }
  return false;
}
////////////////////////////////////////////////////////////////////////
// Debug helper
////////////////////////////////////////////////////////////////////////
function mzDebug(error)
{
  if (gDebug)
    dump("\t *** MultiZilla uninstall: " + error + "\n");
}
